#include "oop.h"
#include "minmax.h"

#include <stdio.h>


oop Stream::nextPutAll(oop aCollection)
{
  iterate(aCollection, iter)
    nextPut(iter.element());
  return this;
}


oop Stream::nextPut(oop anObject)
{
  subclassResponsibility(0);
}


oop Stream::contents(void)
{
  // Answer all of the contents of the receiver.
  subclassResponsibility(0);
}


/// ----------------------------------------------------------------


oop PositionableStream::on(oop aCollection)
{
  mCollection= aCollection;
  mReadLimit= aCollection->size();
  mPosition= 0;
  return reset();
}


oop PositionableStream::reset(void)
{
  mPosition= 0;
  return this;
}


int PositionableStream::atEnd(void)
{
  // Answer whether the receiver can access any more objects.
  return mPosition >= mReadLimit;
}


oop PositionableStream::contents(void)
{
  // Answer with a copy of my collection from 0 to readLimit.
  return mCollection->copyFromTo(0, mReadLimit);
}


/// ----------------------------------------------------------------


ReadStream::ReadStream(oop aCollection)
{
  on(aCollection);
}


oop ReadStream::next(void)
{
  // Answer the next object in the Stream represented by the receiver.
  // Fail if the collection of this stream is not an Array or a
  // String.  Fail if the stream is positioned at its end, or if the
  // position is out of bounds in the collection.
  return atEnd() ? 0 : mCollection->at(mPosition++);
}


/// ----------------------------------------------------------------


WriteStream::WriteStream(oop aCollection)
{
  on(aCollection);
}


oop WriteStream::reset(void)
{
  mReadLimit= ::max(mReadLimit, mPosition);
  return PositionableStream::reset();
}


oop WriteStream::on(oop aCollection)
{
  PositionableStream::on(aCollection);
  mReadLimit= 0;
  mWriteLimit= 0;
  return this;
}


oop WriteStream::nextPut(oop anObject)
{
  // Insert the argument at the next position in the Stream
  // represented by the receiver.  Fail if the collection of this
  // stream is not an Array or a String.  Fail if the stream is
  // positioned at its end, or if the position is out of bounds in the
  // collection.  Fail if the argument is not of the right type for
  // the collection.
  return (mPosition >= mWriteLimit)
    ? pastEndPut(anObject)
    : mCollection->atPut(mPosition++, anObject);
}


oop WriteStream::pastEndPut(oop anObject)
{
  // Grow the collection by creating a new bigger collection and then
  // copy over the contents from the old one.  We grow by doubling the
  // size but the growth is kept between 20 and 1000000.  Finally we
  // put anObject at the current write position.
  int oldSize= mCollection->size();
  oop grownCollection= mCollection->classNew(oldSize + ::min(::max(oldSize, 20), 1000000));
  mCollection= grownCollection->replaceFromToWithStartingAt(0, oldSize, mCollection, 0);
  mWriteLimit= mCollection->size();
  return mCollection->atPut(mPosition++, anObject);
}


oop WriteStream::tab(int n)
{
  while (n--)
    nextPut(Character::Tab);
  return this;
}

oop WriteStream::tab(void)
{
  return tab(1);
}

oop WriteStream::nl(int n)
{
  while (n--)
    nextPut(Character::NL);
  return this;
}

oop WriteStream::nl(void)
{
  return nl(1);
}

oop WriteStream::cr(int n)
{
  while (n--)
    nextPut(Character::CR);
  return this;
}

oop WriteStream::cr(void)
{
  return cr(1);
}

oop WriteStream::space(int n)
{
  while (n--)
    nextPut(Character::Space);
  return this;
}

oop WriteStream::space(void)
{
  return space(1);
}

oop WriteStream::print(oop anObject)
{
  anObject->printOn(this);
  return this;
}


oop WriteStream::vformat(char *fmt, va_list ap)
{
  int c;
  while ((c= *fmt++))
    switch (c)
      {
      case '%':
	{
	  int sign=  0;
	  int left=  1;
	  int pad=  ' ';
	  int width= 0;
	more:
	  switch (c= *fmt++)
	    {
	    case 0:
	      --fmt;
	      nextPut(Character::value('%'));
	      continue;

	    case '+':
	      ++sign;
	      goto more;

	    case '-':
	      left= -1;
	      goto more;

	    case '0':
	      if (!width)
		pad= '0';
	      // fall through...

	    case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
	      width= width * 10 + c - '0';
	      goto more;

	    case 'd':
	      {
		char buf[1024];
		if (sign)
		  {
		    if ('0' == pad)
		      sprintf(buf, "%+0*d", left * width, va_arg(ap, int));
		    else
		      sprintf(buf, "%+*d", left * width, va_arg(ap, int));
		  }
		else
		  {
		    if ('0' == pad)
		      sprintf(buf, "%0*d", left * width, va_arg(ap, int));
		    else
		      sprintf(buf, "%*d", left * width, va_arg(ap, int));
		  }
		nextPutAll(new String(buf));
	      }
	      continue;

	    case 'x':
	      {
		char buf[1024];
		if ('0' == pad)
		  sprintf(buf, "%0*x", left * width, va_arg(ap, int));
		else
		  sprintf(buf, "%*x", left * width, va_arg(ap, int));
		nextPutAll(new String(buf));
	      }
	      continue;

	    case 'o':
	      {
		char buf[1024];
		if ('0' == pad)
		  sprintf(buf, "%0*o", left * width, va_arg(ap, int));
		else
		  sprintf(buf, "%*o", left * width, va_arg(ap, int));
		nextPutAll(new String(buf));
	      }
	      continue;

	    case 'P':
	      va_arg(ap, oop)->printOn(this);
	      continue;

	    case 's':
	      {
		char buf[1024];
		sprintf(buf, "%*s", left * width, va_arg(ap, char *));
		nextPutAll(new String(buf));
	      }
	      continue;

	    case 'c':
	      nextPut(Character::value(va_arg(ap, int)));
	      continue;

	    default:
	      nextPut(Character::value('%'));
	      nextPut(Character::value(c));
	      continue;
	    }
	}
	continue;

      default:
	nextPut(Character::value(c));
	continue;
      }
  return this;
}

oop WriteStream::format(char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vformat(fmt, ap);
  va_end(ap);
  return this;
}


oop WriteStream::contents(void)
{
  mReadLimit= ::max(mReadLimit, mPosition);
  return PositionableStream::contents();
}
